unit GeneralData;
{ General functions. Also handles configs, etc... }

interface

const
	UNTESTED = '?';
    CONTESTANT_INFO_FILE = 'contest.txt';

    TASK_CFG_FILE = 'tasks.cfg';
    CONTESTANT_CFG_FILE = 'contestants.cfg';
    COMPILER_CFG_FILE = 'compiler.cfg';

    TEST_DIR = 'testdir';

	SHORT_LOG_SUFFIX = '-short.txt';
	FULL_LOG_SUFFIX = '-full.txt';

type
	{ Grader job entry }
	TJob = record
    	Index: Integer;
    	Folder, Task: String;
    end;

    { Task description }
    TTask = record
    	Name: String;
        Folder: String;
    end;

    { Results for a task }
    TTaskRes = record
    	Name, Result: String; { Name of task and respective result }
    end;

    { Compiler configuration }
    TCompiler = record
    	Sign, CmdLine: String;
    end;

    { Contestant is mainly defined by his working folder }
    TContestant = class
        private
        	Folder: String;
	        Name, Surname, FatherName, ScClass, School,
          Rayon, Tochka: String;
	        TaskResult: array of TTaskRes;
    	public
			constructor create(const fold: String);

        	function GetFolder: String;
            function Display: String;
            function GetTotalScore: Integer;
            function GetScClass: String;
            function GetSchool: String;
            function GetRayon: String;
            function GetTochka: String;

            procedure LoadTaskRes(const TaskName: String); overload;
            procedure LoadTaskRes(const TaskName: array of String); overload;
            function GetTaskRes(const TaskName: String): String;
            function GetTaskPoints(const TaskName: String): Integer;
            function GetTotalPoints: Integer;
    end;

var
	USER_NAME_FORMAT: String = '$SURNAME $NAME $FATHERNAME';
	JobList: array of TJob;
    ContestantList: array of TContestant;
    TaskList: array of TTask;
    CompilerList: array of TCompiler;
    WorkDir: String;
    UpdateList: array of Integer;

function SilentStrToInt(const S: String): Integer;
procedure ReplaceStg(var S: String; const substr, repl: String);

procedure AddJob(const Job: TJob);
function PopJob(var Job: TJob): Boolean;
procedure AddContestant(Folder: String);
procedure ReloadContestantResults(var Cont: TContestant);
procedure LoadWorks(Folder: String);
function AddTask(const Name, Folder: String): Boolean;
procedure DeleteTask(const Name: String);
procedure ExportResults(const FileName: String);
procedure LoadConfigs;
procedure SaveConfigs;

function GetName(const S: String): String;
function GetExt(const S: String): String;
procedure StoreToFile(const FileName, Value: String);
procedure MakeReport(const Folder, Task, Short, Full: String);

implementation

uses
	Windows, SysUtils, Grade;

{ MakeFoder()
  Adds trailing backslash if necessary. }
function MakeFolder(const S: String): String;
begin
    if S[Length(S)] <> '\' then
    	Result := S + '\' else
        Result := S;
end;

{ SilentStrToInt()
  Function converts string to integer. If failes - returns 0. In case of
  error this function does not raise any exception. }
function SilentStrToInt(const S: String): Integer;
var
	code: Integer;
begin
	val(S, Result, code);
    if code <> 0 then
    	Result := 0; 
end;

{ ReplaceStg()
  Function searches for all occurances of a given substring and replaces it
  by our string. }
procedure ReplaceStg(var S: String; const substr, repl: String);
var
	i: Integer;
begin
	while TRUE do
    begin
    	i := pos(substr, S);
        if i = 0 then
        	Exit; { Not found! }
        Delete(S, i, Length(substr));
        Insert(repl, S, i);
    end;
end;

{ constructor
  It just loads contestant info. }
constructor TContestant.create(const fold: String);
var
	f: TextFile;
    i: Integer;
begin
    Folder := MakeFolder(fold);

	{$I-}
    AssignFile(f, Folder + CONTESTANT_INFO_FILE);
    Reset(f);
    if IOResult = 0 then
    begin
    	//ReadLn(f, Surname);
    	ReadLn(f, Name);
    	//ReadLn(f, FatherName);
      ReadLn(f, Rayon);
      Readln(f, School);
      Readln(f, ScClass);
      ReadLn(f, Tochka);
        CloseFile(f);
    end;
    {$I+}

    SetLength(TaskResult, 0);
    for i := 0 to Length(TaskList)-1 do
        LoadTaskRes(TaskList[i].Name);
end;

{ GetFolder()
  Returns contestant working folder }
function TContestant.GetFolder: String;
begin
	Result := Folder;
end;

{ Display()
  Formats user name for output. }
function TContestant.Display: String;
begin
{	Result := USER_NAME_FORMAT;
    ReplaceStg(Result, '$SURNAME', Surname);
    ReplaceStg(Result, '$NAME', Name);
    ReplaceStg(Result, '$FATHERNAME', FatherName);}
  Result := Name;
end;

{ GetTotalScore()
  How many points does this contestant earn? =) }
function TContestant.GetTotalScore: Integer;
var
	i: Integer;
begin
	Result := 0;
    for i := 0 to Length(TaskList)-1 do
    	Inc(Result, SilentStrToInt(TaskResult[i].Result));
end;

{ GetScClass()
  Returns school class }
function TContestant.GetScClass: String;
begin
	Result := ScClass;
end;

{ GetSchool()
  Returs contestant school }
function TContestant.GetSchool: String;
begin
	Result := School;
end;

{ LoadTaskRes()
  This procedure loads task. If task already loaded - recache }
procedure TContestant.LoadTaskRes(const TaskName: String);
var
	i, ind: Integer;
    f: TextFile;
begin
	{ Maybe we already have an allocated place? }
	ind := -1;
    for i := 0 to Length(TaskResult)-1 do
    	if TaskResult[i].Name = TaskName then
        begin
        	ind := i;
            Break;
        end;

    { Maybe we need to resize our array? }
    if ind = -1 then
    begin
    	ind := Length(TaskResult);
        SetLength(TaskResult, Length(TaskResult)+1);
        TaskResult[ind].Name := TaskName;
    end;

    { Read score }
	AssignFile(f, Folder + TaskName + SHORT_LOG_SUFFIX);
    {$I-}
    Reset(f);
    if IOResult = 0 then
    begin
    	ReadLn(f, TaskResult[ind].Result);
    	CloseFile(f);
    end else
    	TaskResult[ind].Result := UNTESTED;
    {$I+}
end;

{ LoadTaskRes()
  An overloaded variant of previouse procedure. }
procedure TContestant.LoadTaskRes(const TaskName: array of String);
var
	i: Integer;
begin
	for i := 0 to Length(TaskName)-1 do
    	LoadTaskRes( TaskName[i] );
end;

{ GetTaskRes()
  Returns the result for a given task. }
function TContestant.GetTaskRes(const TaskName: String): String;
var
	i: Integer;
begin
	for i := 0 to Length( TaskResult )-1 do
    	if TaskResult[i].Name = TaskName then
        begin
        	Result := TaskResult[i].Result;
        	Exit;
        end;
    Result := '';
end;

{ GetTaskPoints()
  Return's the result for a given task in integer number of points. }
function TContestant.GetTaskPoints(const TaskName: String): Integer;
begin
	Result := SilentStrToInt(TaskName);
end;

{ GetTotalPoints()
  Find total amount of points, earned in a contest }
function TContestant.GetTotalPoints: Integer;
var
	i: Integer;
begin
	Result := 0;
    for i := 0 to Length( TaskResult )-1 do
    	Inc(Result, SilentStrToInt( TaskResult[i].Result ));
end;

{ AddJob()
  This procedure adds test job to the list of jobs. It correctly handles
  multithreading. }
procedure AddJob(const Job: TJob);
var
	GradeID: DWORD;
	i: Integer;
begin
	while JobQueueBusy do
    	Sleep(0);

    JobQueueBusy := True;

    for i := 0 to Length(JobList)-1 do
    	if (JobList[i].Folder = Job.Folder) AND
           (JobList[i].Task = Job.Task) then
        	Exit;

    SetLength(JobList, Length(JobList)+1);
    JobList[ Length(JobList)-1 ] := Job;

	JobQueueBusy := False;

    if not GradeThreadExec then
    begin
		GradeThreadExec := True;
        CreateThread(NIL,   // No security attributes
                     256*1024, // 0.25Mb stack size
                     @GradeThreadProc, // Tread proc
                     NIL,   // No parameters for thread
                     0,     // No additional creation options
                     GradeID); //
	end;
end;

{ PopJob()
  Function gets a job from a queue. If job queue is empty - returns false,
  otherwise - true. This procedure also correctly handles multithreading. }
function PopJob(var Job: TJob): Boolean;
begin
	while JobQueueBusy do
    	Sleep(0);

    JobQueueBusy := True;

    if Length( JobList ) = 0 then
    begin
		Result := False;
        JobQueueBusy := False;
        Exit;
    end;

    Job := JobList[ Length(JobList)-1 ];
    SetLength(JobList, Length(JobList)-1);
	Result := True;
    JobQueueBusy := False;
end;

{ AddContestant()
  Procedure adds contestant to the list of all contestants. }
procedure AddContestant(Folder: String);
var
	i: Integer;
begin
	Folder := MakeFolder(Folder);
    for i := 0 to Length(ContestantList)-1 do
    	if ContestantList[i].GetFolder = Folder then
        	Exit; { Already present }
    SetLength(ContestantList, Length(ContestantList)+1);
	ContestantList[ Length(ContestantList)-1 ] := TContestant.create(Folder);
end;

{ ReloadContestantResults()
  Load result for every task. }
procedure ReloadContestantResults(var Cont: TContestant);
var
	i: Integer;
begin
	for i := 0 to Length( TaskList )-1 do
    	Cont.GetTaskRes( TaskList[i].Name );
end;

{ LoadWorks()
  Recursive procedure to load all contestant's works. }
procedure LoadWorks(Folder: String);
var
	S: String;
    Mask: AnsiString;
	Handle: THandle;
    Data: TWIN32FindData;
begin
	if Folder[ Length(Folder) ] <> '\' then
    	Folder := Folder + '\';

    if FileExists(Folder + CONTESTANT_INFO_FILE) then
        AddContestant(Folder);

    { Recurse through all folders }
	Mask := Folder + '*';
    Handle := FindFirstFile(@Mask[1], Data);
    if Handle = INVALID_HANDLE_VALUE then
    begin
	    RaiseLastOSError;
    	Exit;
    end;
    repeat
    	if Data.dwFileAttributes AND FILE_ATTRIBUTE_DIRECTORY <> 0 then
        begin
        	S := Data.cFileName;
            if (S <> '.') and (S <> '..') then
            	LoadWorks(Folder + S);
        end;
    until not FindNextFile(Handle, Data);
    Windows.FindClose(Handle);
end;

{ AddTask()
  Function adds a task. Function succeeds if no such task already exist }
function AddTask(const Name, Folder: String): Boolean;
var
	i, k: Integer;
begin
   	Result := False;
    k := Length(TaskList);
	for i := 0 to k-1 do
    	if TaskList[i].Name = Name then
        	Exit;

    SetLength(TaskList, k+1);
    TaskList[ k ].Name := Name;
    TaskList[ k ].Folder := Folder;

    { Load results for this task for every contestant }
    for i := 0 to Length( ContestantList )-1 do
    	ContestantList[i].LoadTaskRes(Name);
    Result := True;
end;

{ DeleteTask()
  Procedure erases given task. }
procedure DeleteTask(const Name: String);
var
	i, k: Integer;
begin
	k := 0;
    for i := 0 to Length(TaskList)-1 do
    	if TaskList[i].Name <> Name then
        begin { We can't delete this task }
        	if k <> i then
	        	TaskList[k] := TaskList[i]; { We need to shift }
            Inc(k);
        end;

    if k <> Length(TaskList) then
	    SetLength(TaskList, k);
end;

{ ExportResults()
  This procedure exports results to a given file. }
procedure ExportResults(const FileName: String);
var
	i, j: Integer;
	f: TextFile;
    S: String;
begin
	AssignFile(f, FileName);
    Rewrite(f);
    	WriteLn(f, '<html>');
    	WriteLn(f, '  <head>');
    	WriteLn(f, '    <title> </title>');
    	WriteLn(f, '  </head>');
    	WriteLn(f, '  <body>');
    	WriteLn(f, '    <table border="1" width="100%">');

		{ Draw header }
    	WriteLn(f, '      <tr>');
{    	WriteLn(f, '        <th align="center"></th>');
    	WriteLn(f, '        <th align="center"></th>');
    	WriteLn(f, '        <th align="center"></th>');
    	WriteLn(f, '        <th align="center"></th>');}

    	WriteLn(f, '        <th align="center"></th>');
    	WriteLn(f, '        <th align="center"></th>');
    	WriteLn(f, '        <th align="center"></th>');
    	WriteLn(f, '        <th align="center"></th>');
    	WriteLn(f, '        <th align="center"></th>');
    	WriteLn(f, '        <th align="center"></th>');

        for i := 0 to Length( TaskList )-1 do
	    	WriteLn(f, '        <th align="center">', TaskList[i].Name, '</th>');
    	WriteLn(f, '        <th align="center"></th>');
    	WriteLn(f, '      </tr>');

        { For each contestant draw his result }
        for i := 0 to Length( ContestantList )-1 do
        begin
	    	WriteLn(f, '      <tr>');
    		WriteLn(f, '        <td>', ContestantList[i].Display, '</td>');
    		WriteLn(f, '        <td>', ContestantList[i].GetRayon, '</td>');
    		WriteLn(f, '        <td>', ContestantList[i].GetScClass, '</td>');
    		WriteLn(f, '        <td>', ContestantList[i].GetSchool, '</td>');
    		WriteLn(f, '        <td>', ContestantList[i].GetTochka, '</td>');


            { Working folder }
            S := ContestantList[i].GetFolder;
            Delete(S, Length(S), 1);
            while pos('\', S) <> 0 do
            	Delete(S, 1, pos('\', S));

    		WriteLn(f, '        <td>', S, '</td>');
	        for j := 0 to Length( TaskList )-1 do
		    	WriteLn(f, '        <td>',
                	ContestantList[i].GetTaskRes( TaskList[j].Name ), '</td>');
    		WriteLn(f, '        <td>', ContestantList[i].GetTotalPoints, '</td>');
    		WriteLn(f, '      </tr>');
        end;

    	WriteLn(f, '    </table>');
    	WriteLn(f, '  </body>');
    	WriteLn(f, '</html>');
    CloseFile(f);
end;

{ LoadConfigs()
  Procedure loads task list and contestant list. }
procedure LoadConfigs;
var
	f: TextFile;
    S: String;
    TaskNameList: array of String;
    i, k: Integer;
begin
	SetLength(TaskList, 0);
    SetLength(ContestantList, 0);

{$I-}
	{ Load task list }
    AssignFile(f, WorkDir + '\' + TASK_CFG_FILE);
    Reset(f);
    if IOResult = 0 then
    begin
    	ReadLn(f, k);
        SetLength(TaskList, k);
        for i := 0 to k-1 do
        begin
        	ReadLn(f, S);
            TaskList[i].Name := copy(S, 1, pos('*', S)-1);
            TaskList[i].Folder := copy(S, pos('*', S)+1, Length(S));
        end;
	    CloseFile(f);
    end;

	{ Load contestant list }
    AssignFile(f, WorkDir + '\' + CONTESTANT_CFG_FILE);
    Reset(f);
    if IOResult = 0 then
    begin
    	ReadLn(f, k);
        for i := 1 to k do
		begin
        	ReadLn(f, S);
            AddContestant(S);
		end;
	    CloseFile(f);
    end;

    { Load compiler configs }
    AssignFile(f, WorkDir + '\' + COMPILER_CFG_FILE);
    Reset(f);
    if IOResult = 0 then
    begin
    	ReadLn(f, k);
        SetLength(CompilerList, k);
        for i := 0 to k-1 do
        begin
        	ReadLn(f, CompilerList[i].Sign);
        	ReadLn(f, CompilerList[i].CmdLine);
        end;
    end;
    CloseFile(f);
{$I+}

	{ Get result for each task }
    SetLength(TaskNameList, Length(TaskList) );
    for i := 0 to Length( ContestantList )-1 do
      	ContestantList[i].LoadTaskRes(TaskNameList);
end;

{ SaveConfigs()
  Procedure saves task list and contestant list. }
procedure SaveConfigs;
var
	f: TextFile;
    i: Integer;
begin
{$I-}
	{ Save task list }
    AssignFile(f, WorkDir + '\' + TASK_CFG_FILE);
    Rewrite(f);
    	WriteLn(f, Length(TaskList));
        for i := 0 to Length(TaskList)-1 do
        	WriteLn(f, TaskList[i].Name, '*', TaskList[i].Folder);
    CloseFile(f);

	{ Save contestant list }
    AssignFile(f, WorkDir + '\' + CONTESTANT_CFG_FILE);
    Rewrite(f);
    	WriteLn(f, Length(ContestantList));
        for i := 0 to Length(ContestantList)-1 do
        	WriteLn(f, ContestantList[i].Folder);
    CloseFile(f);

    { Save compiler configs }
    AssignFile(f, WorkDir + '\' + COMPILER_CFG_FILE);
    Rewrite(f);
    	WriteLn(f, Length(CompilerList));
        for i := 0 to Length(CompilerList)-1 do
        begin
        	WriteLn(f, CompilerList[i].Sign);
        	WriteLn(f, CompilerList[i].CmdLine);
        end;
    CloseFile(f);
{$I+}
end;

{ GetName()
  Function parses whole path and returns only file name }
function GetName(const S: String): String;
begin
	Result := S;
	while pos('\', Result) <> 0 do
    	Delete(Result, 1, pos('\', Result));
end;

{ GetExt()
  Function parses extension  }
function GetExt(const S: String): String;
begin
	Result := GetName(S);
	while pos('.', Result) <> 0 do
    	Delete(Result, 1, pos('.', Result));
end;

{ WriteFile()
  Simple procedure to write one string to file }
procedure StoreToFile(const FileName, Value: String);
var
	f: TextFile;
begin
	AssignFile(f, FileName);
    Rewrite(f);
	if Value <> '' then
    	WriteLn(f, Value);
    Close(f);
end;


{ MakeReport()
  This procedure makes report }
procedure MakeReport(const Folder, Task, Short, Full: String);
var
	i: Integer;
begin
	{ Store log to files }
    StoreToFile(Folder + Task + SHORT_LOG_SUFFIX, Short);
    StoreToFile(Folder + Task + FULL_LOG_SUFFIX, Full);

    { Update info }
    for i := 0 to Length( ContestantList )-1 do
    	if ContestantList[i].GetFolder = Folder then
        begin
        	ContestantList[i].LoadTaskRes(Task);
            SetLength( UpdateList, Length(UpdateList)+1 );
            UpdateList[ Length(UpdateList)-1 ] := i;
            Exit;
        end;
end;

function TContestant.GetRayon: String;
begin
  Result := Rayon;
end;

function TContestant.GetTochka: String;
begin
  Result := Tochka;
end;

end.
